"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var stringHash = require("string-hash");
var Declaration = /** @class */ (function () {
    function Declaration(kind, type) {
        this.kind = kind;
        this.type = type;
    }
    Declaration.prototype.equals = function (other) {
        if (!(other instanceof Declaration))
            return false;
        return this.kind === other.kind && this.type.equals(other.type);
    };
    Declaration.prototype.hashCode = function () {
        return (stringHash(this.kind) + this.type.hashCode()) | 0;
    };
    return Declaration;
}());
exports.Declaration = Declaration;
var DeclarationIR = /** @class */ (function () {
    function DeclarationIR(declarations, forwardedTypes) {
        this.declarations = declarations;
        this.forwardedTypes = forwardedTypes;
    }
    return DeclarationIR;
}());
exports.DeclarationIR = DeclarationIR;
function declarationsForGraph(graph, needsForwardDeclarations, childrenOfType, typeNeedsDeclaration) {
    var visitedTypes = immutable_1.Set();
    var forwardedTypes = immutable_1.Set();
    var declarations = [];
    function visit(t, path) {
        if (visitedTypes.has(t))
            return;
        if (path.has(t)) {
            if (needsForwardDeclarations) {
                declarations.push(new Declaration("forward", t));
                forwardedTypes = forwardedTypes.add(t);
            }
            return;
        }
        var pathForChildren = path.add(t);
        childrenOfType(t).forEach(function (c) { return visit(c, pathForChildren); });
        if (visitedTypes.has(t))
            return;
        if (forwardedTypes.has(t) || typeNeedsDeclaration(t)) {
            declarations.push(new Declaration("define", t));
            visitedTypes = visitedTypes.add(t);
        }
    }
    var topLevels = graph.topLevels;
    if (needsForwardDeclarations) {
        topLevels = topLevels.reverse();
    }
    topLevels.forEach(function (t) { return visit(t, immutable_1.Set()); });
    var declarationsList = immutable_1.List(declarations);
    if (!needsForwardDeclarations) {
        declarationsList = declarationsList.reverse();
    }
    return new DeclarationIR(declarationsList, forwardedTypes);
}
exports.declarationsForGraph = declarationsForGraph;
function cycleBreakerTypesForGraph(graph, canBeCycleBreaker) {
    var visitedTypes = immutable_1.Set();
    var cycleBreakerTypes = immutable_1.Set();
    var queue = graph.topLevels.valueSeq().toArray();
    function visit(t, path) {
        if (visitedTypes.has(t))
            return;
        if (canBeCycleBreaker(t)) {
            if (path.has(t)) {
                cycleBreakerTypes = cycleBreakerTypes.add(t);
                return;
            }
            var pathForChildren_1 = path.add(t);
            t.children.forEach(function (c) { return visit(c, pathForChildren_1); });
        }
        else {
            queue.push.apply(queue, t.children.toArray());
        }
        visitedTypes = visitedTypes.add(t);
    }
    for (;;) {
        var maybeType = queue.pop();
        if (maybeType === undefined)
            break;
        visit(maybeType, immutable_1.Set());
    }
    return cycleBreakerTypes;
}
exports.cycleBreakerTypesForGraph = cycleBreakerTypesForGraph;
